#include "SendRecvThread.h"
#include "MySyncObject.h"		// CMySyncObjectを使うため
#include "RingBuff.h"			// CRingBuffを使うため

// 送信速度
#define		SEND_BPS	(1000000.0)	// 送信速度1Mbps

//----------------------------------------------
// function
//    コンストラクタ
// parameter
//    ConnectionInfoRec *pConInfo	[in]機能に必要な情報
// return
//    なし
//----------------------------------------------
CSendRecvThread::CSendRecvThread(ConnectionInfoRec *pConInfo)
{
	m_pConInfo = pConInfo;
	m_fIamZombie = FALSE;
	m_pCRingBuffSend = new CRingBuff(SEND_BUFF_SIZE);	// 送信リングバッファの構築
	m_pCRingBuffRecv = new CRingBuff(RECV_BUFF_SIZE);	// 受信リングバッファの構築

	m_dwPrevSentTime = 0;						// 初回送信はすぐに
	m_dwSendInterval = 0;						// 
}


//----------------------------------------------
// function
//    デストラクタ
// parameter
//    なし
// return
//    なし
//----------------------------------------------
CSendRecvThread::~CSendRecvThread()
{
	SAFE_DELETE(m_pCRingBuffSend)		// 送信リングバッファの破棄
	SAFE_DELETE(m_pCRingBuffRecv)		// 受信リングバッファの破棄
}

//----------------------------------------------
// function
//    機能を記述した関数
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
UINT CSendRecvThread::DoWork()
{
	BOOL		fRet = TRUE;
	pollfd		fds[1] = { 0 };
	char		*pcData = NULL;				// 未送信データ
	int			iSendSize = 0;				// 未送信データサイズ
	char	    szRecvBuffer[RCVBUFSIZE];	// 受信データを一時的に格納
	int			iRecvSize;
	DWORD		dwNow;						// 送信チェックした時刻を覚えるため
	sockaddr_storage	PeerAddr;			// ★送信元アドレス情報
	socklen_t 			iPeerLen;			// ★送信元アドレス情報のサイズ
	char		szPeerAddr[NI_MAXHOST];		// ★相手アドレス,ポート番号をgetnameinfoで取得
	char		szServiceNo[NI_MAXSERV];	// ★

	fds[0].fd = m_pConInfo->fdClient;
	while (!m_fStopFlag)
	{
		// 送信可能時刻かつ未送信データがあるときだけ送信可能検査用fd_setにセットする
		// 送信データがあるかリングバッファを調べる
		iSendSize = m_pCRingBuffSend->GetReadableSize();
		// チェックのために現在時刻を取得する
		dwNow = timeGetTime();
		fds[0].events = POLLIN | POLLRDHUP;					// 受信と相手側からの切断イベントを設定
		// 送信可能時刻かつ未送信データがあるか
		if ((iSendSize > 0) && (CanSendNow(dwNow) == TRUE))
			fds[0].events = POLLOUT | POLLIN | POLLRDHUP;	// 書き込み可能を追加

		poll(fds, 1, 10);
		if (fds[0].revents & POLLERR)			// エラー発生
		{
			DispErrorMsg("Err DoWork");
			fRet = FALSE;
			break;
		}
		if (fds[0].revents & POLLRDHUP)			// 切断発生
		{
			DispErrorMsg("Disconnected");
			fRet = FALSE;
			break;
		}
		if (fds[0].revents & POLLIN)		// 受信データイベント
		{
			// 受信リングバッファに空きがあれば取得する
			// UDPは受信パケットサイズより小さなバッファでrecvfromを呼ぶと残りが破棄されるので
			// 空きが受信パケットサイズより大きいときにrecvfromする
			if(m_pCRingBuffRecv->GetWriteableSize() >= RCVBUFSIZE)
			{
				iPeerLen = sizeof(PeerAddr);		// 送信元アドレス情報を格納する構造体のサイズ
				// 0バイトの受信はありうる
				if ((iRecvSize = recvfrom(m_pConInfo->fdClient, szRecvBuffer, RCVBUFSIZE, 0, (sockaddr *)&PeerAddr, &iPeerLen)) < 0)
				{
					m_pConInfo->pCMySyncObject->Lock();
					DispErrorMsg("Err:recvfrom");
					m_pConInfo->pCMySyncObject->UnLock();
					fRet = FALSE;
					break;
				}
				else
				{
					// 送信元の表示
					getnameinfo((sockaddr *)&PeerAddr, iPeerLen, szPeerAddr, NI_MAXHOST,
						szServiceNo, NI_MAXSERV, NI_NUMERICHOST | NI_NUMERICSERV);
					fprintf(stderr, "recv from Addr:%s Port:%s\n:", szPeerAddr, szServiceNo);
					// 取得したデータすぐに受信リングバッファに書きこむ
					// 受信リングバッファに書き込むのはこのスレッドだけなので、すべて書き込めるはず
					m_pCRingBuffRecv->Write((LPBYTE)szRecvBuffer, iRecvSize);
				}
			}
		}
		// 受信リングバッファに格納されているデータの解析を行う
		// 複数のパケットが格納されている可能性があるので、FD_ISSETの結果とは無関係に
		// 解析を行うようにする
		if (AnalyzeDataRecv() == -1)					// 返値が-1の時がエラー
		{
			m_pConInfo->pCMySyncObject->Lock();
			DispErrorMsg("Err:Packet format");
			m_pConInfo->pCMySyncObject->UnLock();
			fRet = FALSE;
			break;
		}
		// 送信処理
		if (fds[0].revents & POLLOUT)		// 送信可能ならsend実施
		{
			// 送信したいデータの取得(PATH_MTUより小さくなるように取得する)
			iSendSize = GetSendData(&pcData);
			if (sendto(m_pConInfo->fdClient, pcData, iSendSize, 0, m_pConInfo->ai_addr, (int)m_pConInfo->ai_addrlen) != iSendSize)
			{
				DispErrorMsg("Err:sendto\n");
				fRet = FALSE;
				break;
			}
			SAFE_FREE(pcData)
			// 送信時刻と次回送信までの時間をセット
			m_dwPrevSentTime = dwNow;
			m_dwSendInterval = CalcNextSendInterval(iSendSize);
		}
	}
	m_pConInfo->pCMySyncObject->Lock();
	m_fIamZombie = TRUE;
	m_pConInfo->pCMySyncObject->UnLock();
	SAFE_FREE(pcData)
	return((fRet == TRUE) ? 0 : -1);
}

//----------------------------------------------
// function
//		送信データの設定
// parameter
//		char	*pcData	[in]送信データ
//		int		iSize	[in]データ長
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendRecvThread::SetSendData(char *pcData, int iSize)
{
	BOOL	fRet = FALSE;

	// 送信リングバッファに空きがないときは書き込まない
	fRet = m_pCRingBuffSend->Write((LPBYTE)pcData, iSize);
	return(fRet);
}

//----------------------------------------------
// function
//		送信データの取得
// parameter
//		char **ppcData	[in/out]送信データ
// return
//		データ長
//----------------------------------------------
int CSendRecvThread::GetSendData(char **ppcData)
{
	int		iSize = 0;

	// 送信データがあるかリングバッファのデータサイズを調べる
	if ((iSize = m_pCRingBuffSend->GetReadableSize()) > 0)
	{
		// 送信サイズをPATH_MTUより小さくしておく(1024+共通ヘッダなら大丈夫)
		iSize = min(iSize, SENDBUFSIZE);
		*ppcData = (char *)calloc(iSize, sizeof(char));
		iSize = m_pCRingBuffSend->Read((LPBYTE)*ppcData, iSize);
	}
	return(iSize);
}

//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
BOOL CSendRecvThread::IsZombie()
{
	BOOL	fRet;

	m_pConInfo->pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pConInfo->pCMySyncObject->UnLock();
	return(fRet);
}

//----------------------------------------------
// function
//		受信データの解析
//		受信リングバッファに格納されているデータを調べる
// parameter
//		なし
// return
//		0:パケットが揃っていない
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
int CSendRecvThread::AnalyzeDataRecv()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	// データサイズを調べる
	int			iSize = m_pCRingBuffRecv->GetReadableSize();

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;
	// ヘッダ部を借り読み込みする
	m_pCRingBuffRecv->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	// ヘッダ部の解析
	if (memcmp(Header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING)) != 0)
	{
		iRet = -1;							// 識別子が違うのでエラー
		goto L_END;
	}
	wCmd = ntohs(Header.wCommand);
	fprintf(stderr, "CMD:%d\n", wCmd);

	switch (wCmd)
	{
		case CMD_MSG_DATA:
			iRet = RecvMessagePacket(&Header);
			break;
		default:							// 知らないコマンドなのでエラー
			iRet = -1;
			break;
	}
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		メッセージコマンドの受信
// parameter
//		HeaderRec *pHeader	[in]仮読みしたヘッダ
// return
//		0:パケットが揃っていない
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
int CSendRecvThread::RecvMessagePacket(HeaderRec *pHeader)
{
	int			iRet = 0;
	MsgDataRec	*pMsgData = NULL;
	int			iMsgSize, iSize;
	char		*pszMsg = NULL;

	iMsgSize = ntohs(pHeader->wDataLen);
	// データが足りないときは何もしない
	iSize = m_pCRingBuffRecv->GetReadableSize();
	if (iSize < iMsgSize + sizeof(HeaderRec))
		goto L_END;
	// パケット全体を受信しているので読み込みを実施する
	pMsgData = (MsgDataRec *)calloc(iMsgSize + sizeof(HeaderRec), sizeof(BYTE));
	m_pCRingBuffRecv->Read((LPBYTE)pMsgData, iMsgSize + sizeof(HeaderRec));
	// NULLターミネート分を追加して確保
	pszMsg = (char *)calloc(iMsgSize + 1, sizeof(char));
	memcpy(pszMsg, pMsgData->bMsgData, iMsgSize);

	m_pConInfo->pCMySyncObject->Lock();
	fprintf(stderr, "Msg:recv %s\n", pszMsg);
	m_pConInfo->pCMySyncObject->UnLock();
	SAFE_FREE(pMsgData)
	SAFE_FREE(pszMsg)
	iRet = 1;
L_END:
	return(iRet);
}


//----------------------------------------------
// function
//		次の送信までの間隔(msec)を求める
// parameter
//		なし
// return
//		msec
//----------------------------------------------
DWORD CSendRecvThread::CalcNextSendInterval(int iSentSize)
{
	DWORD	dwInterval = 0;
	double	dbBytePerSec = SEND_BPS / 8.0;
	double	dbCountPerSec;

	if (iSentSize == 0)
		goto L_END;
	dbCountPerSec = dbBytePerSec / (double)iSentSize;
	dwInterval = DWORD(1000.0 / dbCountPerSec);

L_END:
	return(dwInterval);
}


//----------------------------------------------
// function
//		今送信可能時刻か
// parameter
//		DWORD dwNow		[in]現在時刻
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendRecvThread::CanSendNow(DWORD dwNow)
{
	BOOL	fRet = FALSE;

	if (GetdwInterval(dwNow, m_dwPrevSentTime) >= m_dwSendInterval)
		fRet = TRUE;

	return(fRet);
}

