#pragma once
#include "ThreadJob.h"
#include "define.h"			// RecvMessagePacketの引数のため

class CMySyncObject;		// このクラスの使用することを宣言
class CRingBuff;			// CRingBuffを使用するため

#define	SEND_BUFF_SIZE		(1024 * 64)					// 送信リングバッファのサイズ
#if 1
#define	SENDBUFSIZE			(1024 + sizeof(HeaderRec))	// 一度に送信するサイズの最大値(共通ヘッダ分を加えてあります)
#else
#define	SENDBUFSIZE			(1024 * 4)					// TCPで送信速度を1Mbpsより上げたいときは断片化処理をTCPのスタックに任せる
														// スタック(send)に渡すサイズを増やす
#endif
#define	RECV_BUFF_SIZE		(1024 * 64)					// 受信リングバッファのサイズ
// ★UDPは受信パケットサイズより小さなバッファでrecvfromを呼ぶと残りが破棄されるのでこのサイズには注意
// sendtoで送信されるサイズより小さくしない
#define	RCVBUFSIZE			(1024 * 4)					// ★一度に読む最大受信サイズ


typedef struct
{
	SOCKET			fdClient;
	CMySyncObject	*pCMySyncObject;	// 同期オブジェクト
// ★送信先アドレス情報(sendtoのパラメータ)
	sockaddr		*ai_addr;
	size_t			ai_addrlen;
}ConnectionInfoRec;


class CSendRecvThread :
	public CThreadJob
{
public:
	CSendRecvThread(ConnectionInfoRec *pConInfo);	// パラメータをコンストラクタで渡す
	~CSendRecvThread();

	// 基底クラスの関数をオーバーライドする
	// C++11で明示的にoverrideを書くことが出来るようになりました
	// 基底クラスの当該関数にvirtualが書いていないとエラーを出してくれます
	UINT DoWork() override;						// DoSendで実施している内容を記述
	BOOL SetSendData(char *pcData, int iSize);	// 送信データの設定			
	BOOL IsZombie();							// このスレッドはゾンビ状態か

private:
	ConnectionInfoRec *m_pConInfo;				// コンストラクタで渡されるパラメータを格納
												// このスレッド実行中領域が確保されていること
	BOOL m_fIamZombie;							// ゾンビ状態かどうかを保持
	CRingBuff *m_pCRingBuffSend;				// 送信データ格納用リングバッファ
	CRingBuff *m_pCRingBuffRecv;				// 受信データ格納用リングバッファ

	DWORD	m_dwPrevSentTime;					// 前回送信した時刻(msec)
	DWORD	m_dwSendInterval;					// 次回送信までの間隔(msec)

	int GetSendData(char **ppcData);			// 送信データの取得
	int AnalyzeDataRecv();						// 引数と返値を変更
	int RecvMessagePacket(HeaderRec *pHeader);	// 引数と返値を変更

	DWORD CalcNextSendInterval(int iSentSize);	// 次回送信までの間隔
	BOOL CanSendNow(DWORD dwNow);				// 送信して良い時刻になったか
};

