#include	"stdThread.h"
#include	"SendRecvThread.h"	// 送受信スレッドを使うため
#include	"MySyncObject.h"	// 同期オブジェクトを使うため
#include	"define.h"
#include    <ifaddrs.h>			// ★
#include    <net/if.h>			// ★


// メニューで使用する文字
#define		CMD_QUIT_CHAR				'q'
#define		CMD_SEND_MSG_CHAR			'm'

#define		SEND_INTERVAL				(5 * 1000)		// ★メッセージ送信間隔

// 関数の宣言
BOOL CreateSocket(LPCSTR szDestAddress, WORD wDestPort, LPCSTR pszIFName);	// ★ソケット作成と設定
BOOL DestroySocket(SOCKET &fd);							// 切断とソケットの破棄
void Stop();											// すべてのソケットを破棄する
void DispMenu();										// メニューの表示
int	 GetKeyString(LPSTR pszString, int iSize);			// キーボード入力文字の取得
BOOL KillZombei();										// ゾンビ状態のスレッドを破棄する
BOOL SendMessagePacket(char *pcData, int iSize);		// メッセージパケットの送信
BOOL CheckAndSendMessagePacket(char *pcData, int iSize);// ★メッセージコマンドが有効なら送信

sockaddr *GetInterfaceInfo(int iFamily, LPCSTR szIFName, DWORD *pdwIndex);	// ★インターフェースインデクス、ユニキャストアドレスの取得
void SetSendSocketOption(SOCKET fdS, int iFamily, LPCSTR szIFName);			// ★送信ソケットの設定

// 変数の宣言
CSendRecvThread		*m_pCSendRecvThread = NULL;
ConnectionInfoRec	*m_pConInfo = NULL;
CMySyncObject		*m_pCMySyncObject = NULL;


int main(int argc, char *argv[])
{
	char	szKeyInBuff[81] = {0};
	LPSTR	pszIFName = NULL;			// ★送信元インターフェース名
	int		iSize = 0;
	DWORD	dwPevTime = 0, dwNow;		// ★前回メッセージを送信した時刻
	DWORD	dwInterval = SEND_INTERVAL;	// ★送信間隔
	int		iCount = 0;					// ★何回目の文字列送信か
	char	szCount[81];				// ★回数を文字列に
	LPBYTE	pbSendMsg = NULL;			// ★送信文字列

//	EnableESC();						// ★動作を確認するためECSシーケンスによる画面制御をOFFに
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();

	// ★起動パラメータチェック 送信元ポートを指定可能に
	if ((argc != 3) && (argc != 4))
	{
		fprintf(stderr, "%s <Multicast IP> <DestPort> [<InterfaceName>]\n", argv[0]);
		goto L_END;
	}
	if (argc == 4)
		pszIFName = argv[3];

	// ★ソケットの作成 ソケットの設定 送受信スレッドを開始する
	if (CreateSocket(argv[1], (WORD)atol(argv[2]), pszIFName) == FALSE)
		goto L_END;
	// メニューの表示
	DispMenu();
	while (1)
	{
		usleep(10 * 1000);					// usleepを入れておく方がCPU負荷のことを考えると良いでしょう
		if (KillZombei() == TRUE)			// 致命的エラーならプログラム終了
			break;
		// キーボードから入力された文字列が'q'なら終了
		// 'm'ならメッセージを入力後は送信する		
		switch (GetKeyString(szKeyInBuff, sizeof(szKeyInBuff)- 1))
		{
			case 0:							// 入力なし、何もしない
				break;
			case CMD_QUIT_CHAR:				// 'q'入力終了
				goto L_END;
			case CMD_SEND_MSG_CHAR:			// 'm'入力szKeyInBuffに文字列が格納されている
				iSize = 0;
				if ((iSize = strlen(szKeyInBuff)) > 0)	// ""で送信停止
				{
					// 送信は送信データをセットするだけ、実際の送信はCSendRecvThreadで実施
					// ★メッセージパケットの送信(メッセージ+回数)
					iCount = 0;
					DISABLE_C4996
					sprintf(szCount, "%d", iCount);
					ENABLE_C4996
					pbSendMsg = (LPBYTE)calloc(iSize + strlen(szCount), sizeof(BYTE));
					memcpy(pbSendMsg, szCount, strlen(szCount));
					memcpy(pbSendMsg + strlen(szCount), szKeyInBuff, iSize);
					SendMessagePacket((char *)pbSendMsg, iSize + (int)strlen(szCount));
					dwPevTime = timeGetTime();	// 送信した時刻を覚える
					++iCount;
					SAFE_FREE(pbSendMsg)
				}
				// メニューの表示
				DispMenu();
				break;
		}
		// ★メッセージがあるとき5秒経過してたら送信
		if (iSize > 0)
		{
			dwNow = timeGetTime();
			if (GetdwInterval(dwNow, dwPevTime) >= dwInterval)
			{
				DISABLE_C4996
				sprintf(szCount, "%d", iCount);
				ENABLE_C4996
				pbSendMsg = (LPBYTE)calloc(iSize + strlen(szCount), sizeof(BYTE));
				memcpy(pbSendMsg, szCount, strlen(szCount));
				memcpy(pbSendMsg + strlen(szCount), szKeyInBuff, iSize);
				SendMessagePacket((char *)pbSendMsg, iSize + (int)strlen(szCount));
				dwPevTime = dwNow;
				++iCount;
				SAFE_FREE(pbSendMsg)
			}
		}
	}

L_END:
	// すべてのソケットの破棄
	Stop();
	m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	return(0);
}


//----------------------------------------------
// function
//		★UDPソケットの作成
//		送受信スレッドを開始する
// parameter
//  LPCSTR      szDestAddress	[in]送信先アドレス(マルチキャストアドレス)
//  WORD        wDestPort		[in]送信先ポート
//  LPCSTR		pszIFName		[in]送信IF名/NULL
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CreateSocket(LPCSTR szDestAddress, WORD wDestPort, LPCSTR pszIFName)
{
	BOOL		fRet = TRUE;
	SOCKET		fd = INVALID_SOCKET;
	addrinfo	hints, *pres = NULL, *pTemp = NULL;
	char		szPort[NI_MAXSERV];

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_DGRAM;			// UDP
	DISABLE_C4996
	sprintf(szPort, "%d", wDestPort);			// ポート番号=サービス
	ENABLE_C4996
	// アドレスを指定してgetaddrinfoを実行
	if (getaddrinfo(szDestAddress, szPort, &hints, &pres) == 0)
	{
		pTemp = pres;
		while (pTemp != NULL)
		{
			// socketを実行
			if ((fd = socket(pTemp->ai_family, pTemp->ai_socktype, pTemp->ai_protocol)) == INVALID_SOCKET)
				goto L_NEXT;
			// ★送信ソケットの設定
			SetSendSocketOption(fd, pTemp->ai_family, pszIFName);
			// socketが作れたのでこのソケット採用することにして抜ける
			fRet = TRUE;
			break;
L_NEXT:
			pTemp = pTemp->ai_next;
		}
	}
	if (fd == INVALID_SOCKET)		// ソケット作成に失敗
	{
		DispErrorMsg("Err:socket");
		fRet = FALSE;
		goto L_END;
	}
	// 送受信スレッドを開始する
	m_pConInfo = (ConnectionInfoRec *)calloc(1, sizeof(ConnectionInfoRec));
	m_pConInfo->fdClient = fd;
	m_pConInfo->ai_addr = (sockaddr *)calloc(pTemp->ai_addrlen, sizeof(BYTE));
	m_pConInfo->ai_addrlen = pTemp->ai_addrlen;
	memcpy(m_pConInfo->ai_addr, pTemp->ai_addr, pTemp->ai_addrlen);
	m_pConInfo->pCMySyncObject = m_pCMySyncObject;
	m_pCSendRecvThread = new CSendRecvThread(m_pConInfo);
	m_pCSendRecvThread->Begin();

L_END:
	if (pres != NULL)
		freeaddrinfo(pres);
	return(fRet);
}

//----------------------------------------------
// function
//    ソケットの破棄
// parameter
//    SOCKET		&fd[in/out]破棄するソケット
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DestroySocket(SOCKET &fd)
{
	if (fd != INVALID_SOCKET)
	{
		fprintf(stderr, "DestroySocket()\n");
		shutdown(fd, SHUT_RDWR);     // 受信も送信も停止
		close(fd);
		fd = INVALID_SOCKET;
	}
	return(TRUE);
}

//----------------------------------------------
// function
//    送受信スレッドの終了、ソケットの破棄
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void Stop()
{
	if (m_pCSendRecvThread != NULL)
	{
		fprintf(stderr, "Stop()\n");
		m_pCSendRecvThread->End();
		m_pCSendRecvThread->WaitForEnd();
		SAFE_DELETE(m_pCSendRecvThread)

		DestroySocket(m_pConInfo->fdClient);
		SAFE_FREE(m_pConInfo->ai_addr)			// ★
		SAFE_FREE(m_pConInfo)
	}
}

//----------------------------------------------
// function
//    メニューの表示
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void DispMenu()
{
	Locate(1, 8, 2);
	fprintf(stderr, "%c:quit %c:メッセージ送信 : ",
		CMD_QUIT_CHAR, CMD_SEND_MSG_CHAR);
}

#if 0
//----------------------------------------------
// function
//    キーボード入力(改行まで)の取得
//		入力がないときはすぐにリターン
// parameter
//    LPSTR		pszString	[in/out]データ格納用バッファ
//    int		iSize		[in]バッファのサイズ
// return
//    0:入力なし CMD_QUIT_CHAR:終了 CMD_SEND_MSG_CHAR:メッセージ送信
//----------------------------------------------
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;

	if (kbhit())
	{
		switch (getchar())
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				fprintf(stderr, "\nInput Msg : ");
				fgets(pszString, iSize, stdin);
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#else
// リモートデバッグではttyじゃないのでkbhitは使えない
// 標準入力は0なのでselectのままでOKですが、練習のためにpollに
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;
	pollfd			fds[1] = {0};

	fds[0].fd = 0;					// 0:標準入力 1:標準出力 2:標準エラー
	fds[0].events = POLLIN;			// Enterキー入力を対象に
	poll(fds, 1, 10);
	if (fds[0].revents & POLLIN)	// ENTERキー入力があった
	{
		memset(pszString, 0, iSize);
		read(0, pszString, iSize);
		switch (pszString[0])
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				while (1)
				{
					poll(fds, 1, 10);
					if (fds[0].revents & POLLIN)
						read(0, pszString, iSize);
					else
						break;
				}
				fprintf(stderr, "\n送信メッセージを入力してください : ");
				memset(pszString, 0, iSize);
				read(0, pszString, iSize);	// ENTERキーまで読み込む
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#endif


//----------------------------------------------
// function
//	終了済みのCSendRecvThreadを破棄する
// parameter
//	なし
// return
//	TRUE:破棄した/FALSE:破棄すべきものがなかった
//----------------------------------------------
BOOL KillZombei()
{
	BOOL	fRet = FALSE;

	if (m_pCSendRecvThread != NULL)
	{
		if (m_pCSendRecvThread->IsZombie() == TRUE)
		{
			m_pCSendRecvThread->End();
			m_pCSendRecvThread->WaitForEnd();
			SAFE_DELETE(m_pCSendRecvThread);

			DestroySocket(m_pConInfo->fdClient);
			SAFE_FREE(m_pConInfo->ai_addr)
			SAFE_FREE(m_pConInfo)
			fRet = TRUE;
		}
	}
	return(fRet);
}

//----------------------------------------------
// function
//		メッセージパケットの送信
// parameter
//		char	*pcData		[in]UTF-8データ(NULLターミネートなし)
//		int		iSize		[in]データサイズ
// retun
//		なし
//----------------------------------------------
BOOL	SendMessagePacket(char *pcData, int iSize)
{
	MsgDataRec	*pMsgData = NULL;
	int			iPacketSize = sizeof(HeaderRec) + iSize;		// パケットサイズ
	BOOL		fRet = FALSE;

	pMsgData = (MsgDataRec *)calloc(iPacketSize, sizeof(BYTE));	// パケット全体のエリアを確保

	memcpy(pMsgData->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pMsgData->header.wCommand = htons(CMD_MSG_DATA);
	pMsgData->header.wDataLen = htons(iSize);

	memcpy(pMsgData->bMsgData, pcData, iSize);
	fRet = m_pCSendRecvThread->SetSendData((char *)pMsgData, iPacketSize);

	SAFE_FREE(pMsgData)
	return(fRet);
}

//----------------------------------------------
// function
//		★メッセージコマンドが有効なら送信
// parameter
//		char	*pcData		[in]UTF-8データ(NULLターミネートなし)/NULL
//		int		iSize		[in]データサイズ
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CheckAndSendMessagePacket(char *pcData, int iSize)
{
	BOOL	fRet = FALSE;

	if ((pcData != NULL) && (iSize > 0))
		fRet = SendMessagePacket(pcData, iSize);
	return(fRet);
}

//-----------------------------------------------------------------
// Function
//		★指定されたインターフェースの情報を取得する
// Parameter
//		int		iFamily		[in]アドレスファミリ
//		LPCSTR	szIFName	[in]インターフェース名
//		DWORD	*pdwIndex	[in/out]インターフェースインデックス
// Return
//		指定条件のインターフェースの最初のユニキャストアドレス
//-----------------------------------------------------------------
sockaddr *GetInterfaceInfo(int iFamily, LPCSTR szIFName, DWORD *pdwIndex)
{
	int		iRet = 0;
	struct ifaddrs *pifaddrs = NULL, *pifa;
	char	szAddr[NI_MAXHOST], szMask[NI_MAXHOST];
	sockaddr	*pAddrRet = NULL;

	*pdwIndex = 0;
	if (szIFName == NULL)
		goto L_END;
	// アダプタリストの取得
	if (getifaddrs(&pifaddrs) == -1)
		goto L_END;
	// アダプタ一覧を出力
	for (pifa = pifaddrs; pifa != NULL; pifa = pifa->ifa_next)
	{
		if (pifa->ifa_addr == NULL)
			continue;
		if ((pifa->ifa_addr->sa_family != AF_INET) && (pifa->ifa_addr->sa_family != AF_INET6))
			continue;

		fprintf(stderr, "Name:%s\n", pifa->ifa_name);
		// ★インターフェース名が指定のものでなければ、次へ
		if (strcmp(pifa->ifa_name, szIFName) != 0)
			continue;
		// ★ファミリが指定のものでなければ次へ
		if (pifa->ifa_addr->sa_family != iFamily)
			continue;

		// 名前からインデックスを取得する
		*pdwIndex = if_nametoindex(pifa->ifa_name);
		fprintf(stderr, "Index:%d\n", *pdwIndex);
		if (pifa->ifa_addr->sa_family == AF_INET)
		{
			inet_ntop(AF_INET, &((struct sockaddr_in *)pifa->ifa_addr)->sin_addr, szAddr, sizeof(szAddr));
			inet_ntop(AF_INET, &((struct sockaddr_in *)pifa->ifa_netmask)->sin_addr, szMask, sizeof(szMask));
			fprintf(stderr, "  IPv4: %s netmask %s\n", szAddr, szMask);
			pAddrRet = (sockaddr *)calloc(1, sizeof(sockaddr_in));
			memcpy(pAddrRet, pifa->ifa_addr, sizeof(sockaddr_in));
		}
		else if (pifa->ifa_addr->sa_family == AF_INET6)
		{
			inet_ntop(AF_INET6, &((struct sockaddr_in6 *)pifa->ifa_addr)->sin6_addr, szAddr, sizeof(szAddr));
			inet_ntop(AF_INET6, &((struct sockaddr_in6 *)pifa->ifa_netmask)->sin6_addr, szMask, sizeof(szMask));
			fprintf(stderr, "  IPv6: %s netmask %s\n", szAddr, szMask);
			pAddrRet = (sockaddr *)calloc(1, sizeof(sockaddr_in6));
			memcpy(pAddrRet, pifa->ifa_addr, sizeof(sockaddr_in6));
		}
		// ★見つかったのでリターンする
		if (pAddrRet != NULL)
			goto L_END;
		fprintf(stderr, "\n");
	}

L_END:
	if (pifaddrs != NULL)
		freeifaddrs(pifaddrs);
	return(pAddrRet);
}

//----------------------------------------------
// function
//	★送信ソケットの設定
//  送信元インターフェース, ループバック, TTL/HopLimitsの設定
// parameter
//  SOCKET		fd			[in]ソケット
//  int			iFamily		[in]アドレスファミリ
//  LPCSTR		szIFName    [in]送信元インターフェース名/NULL
// return
//  TRUE/FALSE
//----------------------------------------------
void SetSendSocketOption(SOCKET fd, int iFamily, LPCSTR szIFName)
{
	sockaddr    *pAddr = NULL;
	int			iRet;
	DWORD		dwFIndex;

	// 1.送信元インターフェースの設定(インターフェース名が指定されたとき)
	if ((pAddr = GetInterfaceInfo(iFamily, szIFName, &dwFIndex)) != NULL)
	{
		if (iFamily == AF_INET)
			iRet = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF, (char *)&(((sockaddr_in *)pAddr)->sin_addr), sizeof(in_addr));
		else if (iFamily == AF_INET6)
			iRet = setsockopt(fd, IPPROTO_IPV6, IPV6_MULTICAST_IF, (char *)&dwFIndex, sizeof(dwFIndex));
	}

	// 2.ループバックの設定(Linuxでは0でループバックなしになります)
	DWORD		dwLoop = 1;
	if (iFamily == AF_INET)
		iRet = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&dwLoop, sizeof(dwLoop));
	else if (iFamily == AF_INET6)
		iRet = setsockopt(fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, (char *)&dwLoop, sizeof(dwLoop));

	// 3.TTL/HopLimitsの設定
	DWORD		dwTTL = 32;
	if (iFamily == AF_INET)
		iRet = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&dwTTL, sizeof(dwTTL));
	else if (iFamily == AF_INET6)
		iRet = setsockopt(fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, (char *)&dwTTL, sizeof(dwTTL));

	SAFE_FREE(pAddr)
}
